//-----------------------------------------------------------------------
// Canthology: software that makes it easy to create anthologies.
//
// Copyright (C) 2011 Ciaran McHale.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------

package org.canthology.canthology;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;

import org.config4j.Configuration;
import org.config4j.ConfigurationException;
import org.config4j.SchemaValidator;


class DocumentInfo
{
	private Configuration		cfg;
	private String				scope;
	private int					diagnosticsLevel;

	private String[]			copyCmds;
	private String[]			buildCmds;
	private String				workingDir;
	private String				rootFileBase;
	private String				docClassName;
	private String[]			docClassOptions;
	private String[]			packageNames;
	private String[]			preamble;
	private String[]			frontMatter;
	private String[]			bookLikeClasses;
	private String[]			mainMatter;
	private String[]			backMatter;
	private String[]			searchPath;
	private String[]			copyFileExtensions;
	private String[]			extraFilesToCopy;
	private String[]			copyFileIncludes;
	private String[]			copyFileExcludes;
	private String[]			subFileIncludes;
	private String[]			subFileExcludes;
	private String[]			subSearchReplacePairs;
	private String[]			emptyList = new String[0];
	private HashSet				copiedFileSet;


	public DocumentInfo(
		Configuration			cfg,
		String					scope,
		int						diagnosticsLevel) throws CanthologyException
	{
		String					fullyScopedName;
		int						i;
		SchemaValidator			antSv = new SchemaValidator();
		StringBuffer			buf;
		ArrayList				schemaList;
		String					schema[] = {
			"classes_that_support_front_main_and_back_matter = list[string]",
			"working_dir    = string",
			"build_commands = list[string]",

			"macro                           = scope",
			"@ignoreEverythingIn macro",

			"root_file                       = scope",
			"root_file.base_name             = string",
			"root_file.documentclass         = scope",
			"root_file.documentclass.name    = string",
			"root_file.documentclass.options = list[string]",
			"root_file.package               = scope",
			"root_file.package.names         = list[string]",
			// "root_file.package.<subscope> = scope"
			// "root_file.package.<subscope>.options = list[string]"
			"root_file.preamble              = list[string]",
			"root_file.front_matter          = list[string]",
			"root_file.main_matter           = list[string]",
			"root_file.back_matter           = list[string]",

			"copy                                              = scope",
			"copy.commands                                     = list[string]",
			"copy.file_extensions                              = list[string]",
			"copy.search_path                                  = list[string]",
			"copy.extra_files_to_copy                          = list[string]",
			"copy.look_for_copy_commands                       = scope",
			"copy.look_for_copy_commands.in_matching_files     = list[string]",
			"copy.look_for_copy_commands.not_in_matching_files = list[string]",

			"substitutions                       = scope",
			"substitutions.in_matching_files     = list[string]",
			"substitutions.not_in_matching_files = list[string]",
			"substitutions.search_replace_pairs  = table[string,search,"
			                                          + "string,replace]",
		};

		this.cfg = cfg;
		this.scope = scope;
		this.diagnosticsLevel = diagnosticsLevel;
		this.copiedFileSet = new HashSet(50);

		//--------
		// Extend the schema with:
		//	"root_file.package.<subscope>         = scope"
		//	"root_file.package.<subscope>.options = list[string]"
		//--------
		schemaList = new ArrayList();
		packageNames = cfg.lookupList(scope, "root_file.package.names");
		packageNames = Util.removeDuplicates(packageNames);
		for (i = 0; i < packageNames.length; i++) {
			schemaList.add("root_file.package." + packageNames[i] + " = scope");
			schemaList.add("root_file.package." + packageNames[i]
							+ ".options = list[string]");
		}
		schema = Util.concatenate(schema, schemaList);

		//--------
		// Parse the schema and use it to validate the configuration file.
		//--------
		antSv.parseSchema(schema);
		antSv.validate(cfg, scope, "", true, Configuration.CFG_SCOPE_AND_VARS);

		//--------
		// Copy remaining config information into instance variables.
		//--------
		bookLikeClasses = cfg.lookupList(scope,
				"classes_that_support_front_main_and_back_matter",
				new String[]{"book", "memoir"});
		workingDir   = cfg.lookupString(scope, "working_dir");
		buildCmds    = cfg.lookupList(scope,   "build_commands");

		rootFileBase = cfg.lookupString(scope, "root_file.base_name");
		docClassName = cfg.lookupString(scope, "root_file.documentclass.name");
		docClassOptions = cfg.lookupList(scope,
										 "root_file.documentclass.options");
		preamble     = cfg.lookupList(scope,   "root_file.preamble");
		frontMatter  = cfg.lookupList(scope,   "root_file.front_matter" );
		mainMatter   = cfg.lookupList(scope,   "root_file.main_matter");
		backMatter   = cfg.lookupList(scope,   "root_file.back_matter");

		copyCmds = cfg.lookupList(scope, "copy.commands");
		copyFileExtensions = cfg.lookupList(scope, "copy.file_extensions");
		searchPath   = cfg.lookupList(scope, "copy.search_path");
		extraFilesToCopy = cfg.lookupList(scope, "copy.extra_files_to_copy");

		copyFileIncludes = cfg.lookupList(scope,
						"copy.look_for_copy_commands.in_matching_files");
		copyFileExcludes = cfg.lookupList(scope,
						"copy.look_for_copy_commands.not_in_matching_files");

		subFileIncludes = cfg.lookupList(scope,
										"substitutions.in_matching_files");
		subFileExcludes = cfg.lookupList(scope,
										"substitutions.not_in_matching_files");
		subSearchReplacePairs = cfg.lookupList(scope,
										"substitutions.search_replace_pairs");
		subSearchReplacePairs = Util.concatenate(
					new String[]{"(ROOT_FILE_BASE_NAME)", rootFileBase},
					subSearchReplacePairs);
	}


	public void process() throws CanthologyException
	{
		String				localFileName;
		String				texFileName;
		String				contents;
		FileWriter			out;
		boolean				ok;
		String				cmd;
		StringBuffer		cmdOutput;
		int					i;

		ensureDirExists(workingDir);

		//--------
		// Create the root ".tex" file and write it to disk.
		//--------
		localFileName = rootFileBase + ".tex";
		texFileName = workingDir + File.separator + localFileName;
		contents = createRootFileContents(localFileName);
		try {
			out = new FileWriter(texFileName);
			out.write(contents.toString());
			out.close();
		} catch(IOException ex) {
			throw new CanthologyException("error writing to file '"
											+ texFileName + "': " + ex);
		}

		//--------
		// Copy extra support files, if any.
		//--------
		for (i = 0; i < extraFilesToCopy.length; i++) {
			copyFile(extraFilesToCopy[i], true);
		}

		//--------
		// Recursively process copy commands in the root ".tex" file
		//--------
		if (shouldLookForCopyCommands(localFileName)) {
			processAllEmbeddedCopyFileCommands(contents.toString());
		}

		//--------
		// Run the build commands.
		//--------
		for (i = 0; i < buildCmds.length; i++) {
			cmd   = buildCmds[i];
			cmd   = Util.replace(cmd, subSearchReplacePairs );
			if (diagnosticsLevel >= 1) {
				System.out.println("Executing '" + cmd + "' ...");
			}
			cmdOutput = new StringBuffer();
			ok = Util.execCommand(cmd, new File(workingDir), "", cmdOutput);
			if (!ok || diagnosticsLevel >= 3) {
				System.out.println("\n" + cmdOutput.toString() + "\n");
			}
			if (!ok) {
				return;
			}
		}
	}


	private String createRootFileContents(String localFileName)
	{
		String				pkgName;
		String[]			options;
		StringBuffer		contents;
		int					i;
		String				matterPrefix;
		String				matterSuffix;

		//--------
		// Ensure Canthology doesn't get into an infinite loop if
		// there is is a command such as \input{rootFile.tex} in
		// the root ".tex" file.
		//--------
		copiedFileSet.add(localFileName);

		//--------
		// If bookLikeClasses does not list the document's class, then
		// we output the \frontmatter, \mainmatter and \backmatter
		// commands as comments.
		//--------
		matterPrefix = "\n";
		matterSuffix = "\n";
		if (!Util.arrayContains(bookLikeClasses, docClassName)) {
			matterPrefix = "\n% ";
			matterSuffix = " % not implemented by the '" + docClassName
						   + "' class\n";
		}

		//--------
		// If we're using the "memoir" document class with the "article"
		// option, then we must not use \frontmatter, \mainmatter and
		// \backmatter.
		//--------
		if (docClassName.equals("memoir")
			&& (Util.arrayContains(docClassOptions, "article")))
		{
			matterPrefix = "\n% ";
			matterSuffix = " % Don't use with the memoir class and its "
						   + "`article' option\n";
		}

		//--------
		// Create the contents of the root ".tex" file
		//--------
		contents = new StringBuffer();
		contents.append("\\documentclass");
		appendOptionList(contents, docClassOptions);
		contents.append("{" + docClassName + "}\n\n");
		for (i = 0; i < packageNames.length; i++) {
			pkgName = packageNames[i];
			options = cfg.lookupList(scope,
									"root_file.package." + pkgName + ".options",
									emptyList);
			contents.append("\\usepackage");
			appendOptionList(contents, options);
			contents.append("{" + pkgName + "}\n");
		}
		if (preamble.length != 0) {
			contents.append("\n");
		}
		for (i = 0; i < preamble.length; i++) {
			contents.append(preamble[i] + "\n");
		}
		contents.append("\n\\begin{document}\n");
		contents.append(matterPrefix  + "\\frontmatter" + matterSuffix);
		for (i = 0; i < frontMatter.length; i++) {
			contents.append(frontMatter[i] + "\n");
		}
		contents.append(matterPrefix  + "\\mainmatter" + matterSuffix);
		for (i = 0; i < mainMatter.length; i++) {
			contents.append(mainMatter[i] + "\n");
		}
		contents.append(matterPrefix  + "\\backmatter" + matterSuffix);
		for (i = 0; i < backMatter.length; i++) {
			contents.append(backMatter[i] + "\n");
		}
		contents.append("\n\\end{document}\n");

		//--------
		// We don't invoke shouldDoSubsInFile() to determine if we
		// should perform substitutions in the root ".tex" file.
		// Instead, we unconditionally perform subsitutions in it.
		//--------
		return Util.replace(contents.toString(), subSearchReplacePairs);
	}


	private void appendOptionList(StringBuffer contents, String[] options)
	{
		int				i;

		if (options.length == 0) {
			return;
		}
		contents.append("[");
		for (i = 0; i < options.length; i++) {
			contents.append(options[i]);
			if (i < options.length-1) {
				contents.append(",");
			}
		}
		contents.append("]");

	}


	private void ensureDirExists(String dirStr) throws CanthologyException
	{
		File				dir;

		dir = new File(dirStr);
		try {
			dir.mkdirs();
			if (!dir.isDirectory()) {
				throw new CanthologyException("could not create directory '"
												+ dir + "'");
			}
		} catch(Exception ex) {
			throw new CanthologyException("could not create directory '"
												+ dir + "': " + ex);
		}
	}


	private void copyFile(String localFileName, boolean errorIfNotFound)
													throws CanthologyException
	{
		String[]			fileVariants;
		String				fromFileName;
		String				toFileName;
		int					i;

		fileVariants = findFileVariants(localFileName);
		if (fileVariants.length == 0 && errorIfNotFound) {
			throw new CanthologyException("cannot find file '" + localFileName
											+ "'");
		}
		if (fileVariants.length == 0) {
			if (diagnosticsLevel >= 2) {
				System.out.println("Cannot find file '" + localFileName
								   + "': ignoring");
			}
			return;
		}
		for (i = 0; i < fileVariants.length; i += 2) {
			fromFileName = fileVariants[i + 0];
			toFileName   = fileVariants[i + 1];
			copyFileVariant(fromFileName, toFileName);
		}
	}

	private void copyFileVariant(
		String				fromFileName,
		String				toFileName)
	{
		String				localFileName;
		boolean				shouldDoSubs;
		boolean				shouldDoCopyCmds;
		String				contents;

		//--------
		// Ensure we don't copy a file repeatedly. This check
		// avoids redundant copies. It also avoids infinite
		// loops due to (mutually) recursive embedded support
		// file commands.
		//--------
		localFileName = toFileName;
		if (copiedFileSet.contains(localFileName)) {
			return;
		}
		copiedFileSet.add(toFileName);

		toFileName = workingDir + File.separator + toFileName;
		ensureDirExists(new File(toFileName).getParent());

		if (diagnosticsLevel >= 2) {
			System.out.println("Copying '" + fromFileName + "' to '"
							   + toFileName + "'");
		}
		shouldDoSubs     = shouldDoSubsInFile(localFileName);
		shouldDoCopyCmds = shouldLookForCopyCommands(localFileName);
		if (shouldDoSubs || shouldDoCopyCmds) {
			//--------
			// Perform a global search-and-replace on the file's contents,
			// and write the result into the destination file.
			// Also, process any embedded copy file commands.
			//--------
			contents = Util.readTextFile(fromFileName);
			if (shouldDoSubs) {
				contents = Util.replace(contents, subSearchReplacePairs);
			}
			Util.writeTextFile(toFileName, contents);
			if (shouldDoCopyCmds) {
				processAllEmbeddedCopyFileCommands(contents);
			}
		} else {
			Util.copyBinaryFile(fromFileName, toFileName);
		}
	}


	private void processAllEmbeddedCopyFileCommands(String str)
	{
		int					i;

		for (i = 0; i < copyCmds.length; i++) {
			processEmbeddedCopyFileCommand(str, copyCmds[i]);
		}
	}


	private void processEmbeddedCopyFileCommand(
		String				str,
		String				cmd)
	{
		int					fromIndex;
		int					toIndex;
		int					index;
		int					cmdLen;
		int					contentsEndIndex;
		String				fileName;
		char				ch;
		boolean				foundArg;

		contentsEndIndex = str.length() - 1;
		cmdLen = cmd.length();
		fromIndex = 0;
		while ((index = str.indexOf(cmd, fromIndex)) != -1) {
			index += cmdLen;
			if (index == contentsEndIndex) { break; }
			while (index < contentsEndIndex && str.charAt(index) == '[') {
				index = skipOverOptionalArg(str, index);
			}
			if (index == contentsEndIndex) { break; }
			ch = str.charAt(index);
			if (ch != '{') {
				fromIndex = index;
				continue;
			}
			toIndex = str.indexOf("}", index + 1);
			if (toIndex == -1) {
				break;
			}
			fileName = str.substring(index + 1, toIndex);
			copyCommaSeparatedFileList(fileName);
			fromIndex = toIndex + 1;
			if (fromIndex >= contentsEndIndex) {
				break;
			}
		}
	}

	void copyCommaSeparatedFileList(String commaSeparatedFileNames)
	{
		String[]			fileNames;
		int					i;

		fileNames = commaSeparatedFileNames.split(",");
		for (i = 0; i < fileNames.length; i++) {
			copyFile(fileNames[i], false);
		}
	}


	private int skipOverOptionalArg(
		String				str,
		int					indexOfOpenBracket)
	{
		int					strEndIndex;
		int					index;
		char				ch;

		strEndIndex = str.length() - 1;
		Util.assertion(str.charAt(indexOfOpenBracket) == '[');
		index = indexOfOpenBracket + 1;
		while (index < strEndIndex) {
			switch (str.charAt(index)) {
			case '\n':
				return index;
			case ']':
				return index + 1;
			}
			index++;
		}
		return strEndIndex;
	}


	private boolean shouldLookForCopyCommands(String localFileName)
	{
		return seekPermission(localFileName, copyFileIncludes,copyFileExcludes);
	}


	private boolean shouldDoSubsInFile(String localFileName)
	{
		return seekPermission(localFileName, subFileIncludes,subFileExcludes);
	}


	private boolean seekPermission(
		String				str,
		String[]			yesPatterns,
		String[]			noPatterns)
	{
		int					i;
		String				pattern;

		for (i = 0; i < noPatterns.length; i++) {
			pattern = noPatterns[i];
			if (Configuration.patternMatch(str, pattern)) {
				return false;
			}
		}

		for (i = 0; i < yesPatterns.length; i++) {
			pattern = yesPatterns[i];
			if (Configuration.patternMatch(str, pattern)) {
				return true;
			}
		}
		return false;

	}


	private String[] findFileVariants(String localFileName)
	{
		String				dir;
		String				fileName;
		String				extFileName;
		String				ext;
		ArrayList			result = new ArrayList();
		int					spIndex;
		int					cfeIndex;
		File				file;

		for (spIndex = 0; spIndex < searchPath.length; spIndex++) {
			dir = searchPath[spIndex];
			fileName = dir + File.separator + localFileName;
			file = new File(fileName);
			if (file.canRead()) {
				result.add(fileName);
				result.add(localFileName);
			}
			for (cfeIndex = 0; cfeIndex < copyFileExtensions.length; cfeIndex++)
			{
				ext = copyFileExtensions[cfeIndex];
				extFileName = fileName + ext;
				file = new File(extFileName);
				if (file.canRead()) {
					result.add(extFileName);
					result.add(localFileName + ext);
				}
			}
		}
		return (String[])result.toArray(new String[0]);
	}
}

